# Copyright 2022 The GPflow Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Code for specifying and evaluating boolean expressions.
"""
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Any, Callable, Mapping, Tuple, cast

from .argument_ref import ArgumentRef
from .error_contexts import (
    ErrorContext,
    ObjectValueContext,
    ParallelContext,
    StackContext,
)


class ParsedBoolSpec(ABC):
    """ A boolean expression. """

    @abstractmethod
    def get(self, arg_map: Mapping[str, Any], context: ErrorContext) -> Tuple[bool, ErrorContext]:
        """ Evaluate this boolean value. """


def _paren_repr(spec: ParsedBoolSpec) -> str:
    """
    Return the `repr` of `spec`, wrapping it in parenthesis if it is a non-trivial expression.
    """
    result = repr(spec)
    if isinstance(spec, ParsedArgumentRefBoolSpec) and spec.bool_test == BoolTest.BOOL:
        return result
    return f"({result})"


@dataclass(frozen=True)
class ParsedOrBoolSpec(ParsedBoolSpec):
    """ An "or" expression. """

    left: ParsedBoolSpec
    right: ParsedBoolSpec

    def get(self, arg_map: Mapping[str, Any], context: ErrorContext) -> Tuple[bool, ErrorContext]:
        left_value, left_context = self.left.get(arg_map, context)
        right_value, right_context = self.right.get(arg_map, context)
        return (left_value or right_value), ParallelContext((left_context, right_context))

    def __repr__(self) -> str:
        return f"{_paren_repr(self.left)} or {_paren_repr(self.right)}"


@dataclass(frozen=True)
class ParsedAndBoolSpec(ParsedBoolSpec):
    """ An "and" expression. """

    left: ParsedBoolSpec
    right: ParsedBoolSpec

    def get(self, arg_map: Mapping[str, Any], context: ErrorContext) -> Tuple[bool, ErrorContext]:
        left_value, left_context = self.left.get(arg_map, context)
        right_value, right_context = self.right.get(arg_map, context)
        return (left_value and right_value), ParallelContext((left_context, right_context))

    def __repr__(self) -> str:
        return f"{_paren_repr(self.left)} and {_paren_repr(self.right)}"


@dataclass(frozen=True)
class ParsedNotBoolSpec(ParsedBoolSpec):
    """ A "not" expression. """

    right: ParsedBoolSpec

    def get(self, arg_map: Mapping[str, Any], context: ErrorContext) -> Tuple[bool, ErrorContext]:
        right_value, right_context = self.right.get(arg_map, context)
        return (not right_value), right_context

    def __repr__(self) -> str:
        return f"not {_paren_repr(self.right)}"


class BoolTest(Enum):
    """
    Strategies for converting a value to ``bool``.
    """

    BOOL = (bool, lambda arg: arg)
    IS_NONE = ((lambda x: x is None), lambda arg: f"{arg} is None")
    IS_NOT_NONE = ((lambda x: x is not None), lambda arg: f"{arg} is not None")

    @property
    def to_bool(self) -> Callable[[Any], bool]:
        return cast(Callable[[Any], bool], self.value[0])

    @property
    def repr(self) -> Callable[[str], str]:
        return cast(Callable[[str], str], self.value[1])


@dataclass(frozen=True)
class ParsedArgumentRefBoolSpec(ParsedBoolSpec):
    """ A reference to an input argument. """

    argument_ref: ArgumentRef
    bool_test: BoolTest

    def get(self, arg_map: Mapping[str, Any], context: ErrorContext) -> Tuple[bool, ErrorContext]:
        ((arg_value, relative_context),) = self.argument_ref.get(arg_map, context)
        bool_value = self.bool_test.to_bool(arg_value)
        return bool_value, StackContext(relative_context, ObjectValueContext(arg_value))

    def __repr__(self) -> str:
        return self.bool_test.repr(repr(self.argument_ref))



# Copyright 2022 The GPflow Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
A library for annotating and checking the shapes of tensors.
"""
from .accessors import get_check_shapes
from .base_types import Dimension, Shape
from .checker import ShapeChecker
from .checker_context import check_shape, get_shape_checker
from .config import (
    DocstringFormat,
    ShapeCheckingState,
    disable_check_shapes,
    get_enable_check_shapes,
    get_enable_function_call_precompute,
    get_rewrite_docstrings,
    set_enable_check_shapes,
    set_enable_function_call_precompute,
    set_rewrite_docstrings,
)
from .decorator import check_shapes
from .error_contexts import ErrorContext
from .inheritance import inherit_check_shapes
from .integration.tf import install_tf_integration
from .integration.tfp import install_tfp_integration
from .integration.torch import install_torch_integration
from .shapes import get_shape, register_get_shape

install_tf_integration()
install_tfp_integration()
install_torch_integration()

__version__ = "1.0.0"

__all__ = [
    "Dimension",
    "DocstringFormat",
    "ErrorContext",
    "Shape",
    "ShapeChecker",
    "ShapeCheckingState",
    "accessors",
    "argument_ref",
    "base_types",
    "bool_specs",
    "check_shape",
    "check_shapes",
    "checker",
    "checker_context",
    "config",
    "decorator",
    "disable_check_shapes",
    "error_contexts",
    "exceptions",
    "get_check_shapes",
    "get_enable_check_shapes",
    "get_enable_function_call_precompute",
    "get_rewrite_docstrings",
    "get_shape",
    "get_shape_checker",
    "inherit_check_shapes",
    "inheritance",
    "parser",
    "register_get_shape",
    "set_enable_check_shapes",
    "set_enable_function_call_precompute",
    "set_rewrite_docstrings",
    "shapes",
    "specs",
]



# Copyright 2022 The GPflow Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=import-outside-toplevel

from ..base_types import Shape
from ..error_contexts import ErrorContext
from ..shapes import register_get_shape


def install_torch_integration() -> bool:
    """
    Install various hooks to support PyTorch.

    If PyTorch is not installed in this environment this function does nothing.

    :return: Whether PyTorch support hooks actually were installed.
    """
    try:
        import torch
    except ImportError:
        return False  # PyTorch not installed - don't install integrations.

    @register_get_shape(torch.Tensor)
    def get_torch_shape(shaped: torch.Tensor, context: ErrorContext) -> Shape:
        return tuple(shaped.shape)

    return True



# Copyright 2022 The GPflow Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Code for extracting shapes from object.
"""
import inspect
from typing import Any, Callable, Dict, Sequence, Type

from .base_types import Shape
from .error_contexts import ErrorContext, IndexContext, ObjectTypeContext, StackContext
from .exceptions import NoShapeError

GetShape = Callable[[Any, ErrorContext], Shape]


_GET_SHAPE_SENTINEL = object()
_GET_SHAPES: Dict[Type[Any], GetShape] = {}


def register_get_shape(shape_type: Type[Any]) -> Callable[[GetShape], GetShape]:
    """
    Register a function for extracting the shape from a given type of objects.

    Example:

    .. literalinclude:: /examples/test_check_shapes_examples.py
       :start-after: [custom_type]
       :end-before: [custom_type]
       :dedent:

    See also :func:`get_shape`.

    :param shape_type: Type of objects to extract shapes from.
    """
    # Yes, what's happening here looks extremely much like `functools.singledispatch`;
    # however we cannot actually use `functools.singledispatch`, because it uses a for/else
    # statement, which TensorFlow doesn't know how to compile...

    def _register(getter: GetShape) -> GetShape:
        _GET_SHAPES[shape_type] = getter
        return getter

    return _register


def get_shape(shaped: Any, context: ErrorContext) -> Shape:
    """
    Returns the shape of the given object.

    See also :func:`register_get_shape`.

    :param shaped: The objects whose shape to extract.
    :param context: Context we are getting the shape in, for improved error messages.
    :returns: The shape of ``shaped``, or ``None`` if the shape exists, but is unknown.
    :raises NoShapeError: If objects of this type does not have shapes.
    """
    for t in inspect.getmro(shaped.__class__):
        getter = _GET_SHAPES.get(t)
        if getter is not None:
            return getter(shaped, context)

    shape = getattr(shaped, "shape", _GET_SHAPE_SENTINEL)

    if shape == _GET_SHAPE_SENTINEL:
        raise NoShapeError(StackContext(context, ObjectTypeContext(shaped)))

    assert shape is None or isinstance(shape, tuple), shape

    return shape


@register_get_shape(bool)
@register_get_shape(int)
@register_get_shape(float)
@register_get_shape(str)
def get_scalar_shape(shaped: Any, context: ErrorContext) -> Shape:
    return ()


@register_get_shape(list)
@register_get_shape(tuple)
def get_sequence_shape(shaped: Sequence[Any], context: ErrorContext) -> Shape:
    if len(shaped) == 0:
        # If the sequence doesn't have any elements we cannot use the first element to determine the
        # shape, and the shape is unknown.
        return None
    child_shape = get_shape(shaped[0], StackContext(context, IndexContext(0)))
    if child_shape is None:
        return None
    return (len(shaped),) + child_shape



# Copyright 2022 The GPflow Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utilities for setting and getting `check_shapes`
"""
from typing import Any, Callable, Optional

from .base_types import C


def set_check_shapes(func: Callable[..., Any], check_shapes: Callable[[C], C]) -> None:
    """
    Store ``check_shapes`` in ``func``, so that we later can tell which ``check_shapes`` was applied
    to it.
    """
    setattr(func, "__check_shapes__", check_shapes)


def maybe_get_check_shapes(func: Callable[..., Any]) -> Optional[Callable[[C], C]]:
    """
    Get the ``check_shapes`` that was applied to ``func``.

    :returns: The ``check_shapes`` that is wrapping ``func``, and ``None`` if no ``check_shapes`` is
        not wrapping ``func``.
    """
    return getattr(func, "__check_shapes__", None)  # type: ignore[no-any-return]


def get_check_shapes(func: Callable[..., Any]) -> Callable[[C], C]:
    """
    Get the ``check_shapes`` that was applied to ``func``.

    :raises ValueError: If no ``check_shapes`` was applied to ``func``.
    """
    result = maybe_get_check_shapes(func)
    if result is None:
        raise ValueError(f"{func.__name__} does not have a `check_shapes`.")
    return result



# Copyright 2022 The GPflow Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=import-outside-toplevel

import inspect
from typing import Any, Callable

from ..base_types import Shape
from ..config import add_is_compiled_mode
from ..decorator import WrapperPostProcessor, add_wrapper_post_processor
from ..error_contexts import ErrorContext
from ..shapes import register_get_shape


def install_tf_integration() -> bool:
    """
    Install various hooks to support TensorFlow.

    If TensorFlow is not installed in this environment this function does nothing.

    :return: Whether TensorFlow support hooks actually were installed.
    """
    try:
        import tensorflow as tf
    except ImportError:
        return False  # TensorFlow not installed - don't install integrations.

    ########################################
    # Support for getting shapes of tf types
    ########################################

    @register_get_shape(tf.Tensor)
    @register_get_shape(tf.Variable)
    def get_tensorflow_shape(shaped: Any, context: ErrorContext) -> Shape:
        shape = shaped.shape
        if not shape:
            return None
        return tuple(shape)

    ################################################
    # Support for ShapeCheckingState.EAGER_MODE_ONLY
    ################################################

    add_is_compiled_mode(tf.inside_function)

    ########################################################
    # Work-around TensorFlow's many problems with decorators
    ########################################################

    class TfWrapperPostProcessor(WrapperPostProcessor):
        def on_wrap(
            self,
            func: Callable[..., Any],
            wrapped: Callable[..., Any],
            signature: inspect.Signature,
            tf_decorator: bool = False,
        ) -> Callable[..., Any]:
            # Work-around for TensorFlow saved_model expecting methods to have a `self` argument:
            if "self" in signature.parameters:

                wrapped_function = wrapped

                def wrapped_method(self: Any, *args: Any, **kwargs: Any) -> Any:
                    return wrapped_function(self, *args, **kwargs)

                wrapped = wrapped_method

            # Make TensorFlow understand our decoration:
            if tf_decorator:
                tf.compat.v1.flags.tf_decorator.make_decorator(func, wrapped)

            return wrapped

    add_wrapper_post_processor(TfWrapperPostProcessor())

    return True



# Copyright 2022 The GPflow Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=broad-except

"""
Exceptions generated by `check_shapes`.

These rely heavily on the infrastructure in `error_contexts.py`.
"""
from .error_contexts import ErrorContext, MessageBuilder


class CheckShapesError(Exception):
    """
    Common super class for `check_shapes` errors.
    """

    error_message: str

    def __init__(self, context: ErrorContext) -> None:
        builder = MessageBuilder()
        builder.add_line("")
        builder.add_line(self.error_message)
        with builder.indent() as b:
            context.print(b)
        super().__init__(builder.build())

        self.context = context

        # Prevent Keras from rewriting our exception:
        self._keras_call_info_injected = True


class VariableTypeError(CheckShapesError):
    """
    Error raised if a variable is used both as a rank-1 and a variable-rank variable.
    """

    error_message = (
        "Cannot use the same variable to bind both a single dimension"
        " and a variable number of dimensions."
    )


class SpecificationParseError(CheckShapesError):
    """
    Error raised if there was an error parsing the shape specification.
    """

    error_message = "Unable to parse shape specification."


class DocstringParseError(CheckShapesError):
    """
    Error raised if there was an error parsing the shape specification.
    """

    error_message = "Unable to parse docstring."


class ArgumentReferenceError(CheckShapesError):
    """
    Error raised if the argument to check the shape of could not be resolved.
    """

    error_message = "Unable to resolve argument / missing argument."


class ShapeMismatchError(CheckShapesError):
    """
    Error raised if a function is called with tensors of the wrong shape.
    """

    error_message = "Tensor shape mismatch."


class NoShapeError(CheckShapesError):
    """
    Error raised if we are trying to get the shape of an object that does not have a shape.
    """

    error_message = "Unable to determine shape of object."



# Copyright 2022 The GPflow Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Code for specifying expectations around shapes.
"""
from dataclasses import dataclass
from typing import Optional, Tuple

from .argument_ref import ArgumentRef
from .bool_specs import ParsedBoolSpec


@dataclass(frozen=True)
class ParsedNoteSpec:
    note: str

    def __repr__(self) -> str:
        return f"# {self.note}"


@dataclass(frozen=True)
class ParsedDimensionSpec:
    constant: Optional[int]
    variable_name: Optional[str]
    variable_rank: bool
    broadcastable: bool

    def __post_init__(self) -> None:
        assert (
            self.variable_name is None or self.constant is None
        ), "Dimension cannot be both constant and variable."
        if self.variable_rank:
            assert self.constant is None, "Constants cannot have a variable rank."

    def __repr__(self) -> str:
        tokens = []

        if self.broadcastable:
            tokens.append("broadcast ")

        if self.constant is not None:
            tokens.append(str(self.constant))
        elif self.variable_name:
            tokens.append(self.variable_name)
        else:
            if not self.variable_rank:
                tokens.append(".")

        if self.variable_rank:
            tokens.append("...")

        return "".join(tokens)


@dataclass(frozen=True)
class ParsedShapeSpec:
    dims: Tuple[ParsedDimensionSpec, ...]

    def __repr__(self) -> str:
        dims = [repr(dim) for dim in self.dims]
        return f"[{', '.join(dims)}]"


@dataclass(frozen=True)
class ParsedTensorSpec:
    shape: ParsedShapeSpec
    note: Optional[ParsedNoteSpec]

    def __repr__(self) -> str:
        note_str = f"  {self.note}" if self.note is not None else ""
        return f"{self.shape}{note_str}"


@dataclass(frozen=True)
class ParsedArgumentSpec:
    argument_ref: ArgumentRef
    tensor: ParsedTensorSpec
    condition: Optional[ParsedBoolSpec]

    def __repr__(self) -> str:
        tokens = []
        tokens.append(f"{self.argument_ref}: ")
        tokens.append(repr(self.tensor.shape))

        if self.condition is not None:
            tokens.append(" if ")
            tokens.append(repr(self.condition))

        if self.tensor.note is not None:
            tokens.append("  ")
            tokens.append(repr(self.tensor.note))

        return "".join(tokens)


@dataclass(frozen=True)
class ParsedFunctionSpec:
    arguments: Tuple[ParsedArgumentSpec, ...]
    notes: Tuple[ParsedNoteSpec, ...]

    def __repr__(self) -> str:
        lines = [repr(argument) for argument in self.arguments] + [
            repr(note) for note in self.notes
        ]
        return "\n".join(lines)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['check_shapes', 'check_shapes.integration']

package_data = \
{'': ['*']}

install_requires = \
['lark>=1.1.0,<2.0.0']

setup_kwargs = {
    'name': 'check-shapes',
    'version': '1.0.0',
    'description': 'A library for annotating and checking the shapes of tensors.',
    'long_description': '# check_shapes\n\n`check_shapes` is a library for annotating and checking tensor shapes.\nFor example:\n\n```python\nimport tensorflow as tf\n\nfrom gpflow.experimental.check_shapes import check_shapes\n\n@tf.function\n@check_shapes(\n    "features: [batch..., n_features]",\n    "weights: [n_features]",\n    "return: [batch...]",\n)\ndef linear_model(features: tf.Tensor, weights: tf.Tensor) -> tf.Tensor:\n    return tf.einsum("...i,i -> ...", features, weights)\n```\n\nFor more information see our [documentation](https://gpflow.github.io/check_shapes).\n\n## Installation\n\nThe recommended way to install `check_shapes` is from pypi:\n\n```bash\npip install check_shapes\n```\n\n### From source\n\nTo develop `check_shapes`, check it out from GitHub:\n\n```bash\ngit clone git@github.com:GPflow/check_shapes.git\n```\n\nWe use [Poetry](https://python-poetry.org/) to install and manage dependencies. Follow their\ninstructions for how to install Poetry itself. Then:\n\n```bash\ncd check_shapes\npoetry install\n```\n\nTo check you installation run our tests:\n\n```bash\npoetry run task test\n```\n\nFor testing with different versions of Python and dependencies, see the `poetryenv` script.',
    'author': 'Jesper Nielsen',
    'author_email': 'jespernielsen1982+check_shapes@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://gpflow.github.io/check_shapes',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



# Copyright 2022 The GPflow Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Code for (de)referencing arguments.
"""
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Iterable, List, Mapping, Sequence, Tuple

from .error_contexts import (
    ArgumentContext,
    AttributeContext,
    ErrorContext,
    IndexContext,
    MappingKeyContext,
    MappingValueContext,
    StackContext,
)
from .exceptions import ArgumentReferenceError

# The special name used to represent the returned value. `return` is a good choice because we know
# that no argument can be called `return` because `return` is a reserved keyword.
RESULT_TOKEN = "return"


class ArgumentRef(ABC):
    """ A reference to an argument. """

    @property
    @abstractmethod
    def is_result(self) -> bool:
        """ Whether this is a reference to the function result. """

    @property
    @abstractmethod
    def root_argument_name(self) -> str:
        """
        Name of the argument this reference eventually starts from.

        Returns `RESULT_TOKEN` if this in an argument to the function result.
        """

    @abstractmethod
    def get(
        self, arg_map: Mapping[str, Any], context: ErrorContext
    ) -> Sequence[Tuple[Any, ErrorContext]]:
        """
        Get the value(s) of this argument from the given argument map.
        """

    @abstractmethod
    def __repr__(self) -> str:
        """ Return a string representation of this reference. """


@dataclass(frozen=True)
class RootArgumentRef(ArgumentRef):
    """ A reference to a single argument. """

    argument_name: str

    @property
    def is_result(self) -> bool:
        return self.argument_name == RESULT_TOKEN

    @property
    def root_argument_name(self) -> str:
        return self.argument_name

    def get(
        self, arg_map: Mapping[str, Any], context: ErrorContext
    ) -> Sequence[Tuple[Any, ErrorContext]]:
        relative_context = ArgumentContext(self.argument_name)

        try:
            arg_value = arg_map[self.argument_name]
        except Exception as e:
            raise ArgumentReferenceError(StackContext(context, relative_context)) from e

        return [(arg_value, relative_context)]

    def __repr__(self) -> str:
        return self.argument_name


@dataclass(frozen=True)  # type: ignore[misc]
class DelegatingArgumentRef(ArgumentRef):
    """ Abstract base class for :class:`ArgumentRef`\\ s that delegates to a source. """

    source: ArgumentRef

    @property
    def is_result(self) -> bool:
        return self.source.is_result

    @property
    def root_argument_name(self) -> str:
        return self.source.root_argument_name

    @abstractmethod
    def map_value(self, value: Any, context: ErrorContext) -> Iterable[Tuple[Any, ErrorContext]]:
        """
        Map this value, from `self.source` to new value(s).
        """

    def map_context(self, context: ErrorContext) -> ErrorContext:
        """
        Pre-map this error context from `self.source`.

        The mapped value will both be used for error messages and passed to `map_value` above.
        """
        return context

    def get(
        self, arg_map: Mapping[str, Any], context: ErrorContext
    ) -> Sequence[Tuple[Any, ErrorContext]]:
        results: List[Tuple[Any, ErrorContext]] = []
        sources = self.source.get(arg_map, context)
        for source, source_relative_context in sources:

            if source is None:
                results.append((source, source_relative_context))
                continue

            try:
                relative_context = self.map_context(source_relative_context)
            except Exception as e:
                raise ArgumentReferenceError(context) from e

            try:
                results.extend(self.map_value(source, relative_context))
            except Exception as e:
                raise ArgumentReferenceError(StackContext(context, relative_context)) from e

        return results


@dataclass(frozen=True)
class AttributeArgumentRef(DelegatingArgumentRef):
    """ A reference to an attribute on an argument. """

    attribute_name: str

    def map_value(self, value: Any, context: ErrorContext) -> Iterable[Tuple[Any, ErrorContext]]:
        return [(getattr(value, self.attribute_name), context)]

    def map_context(self, context: ErrorContext) -> ErrorContext:
        return StackContext(context, AttributeContext(self.attribute_name))

    def __repr__(self) -> str:
        return f"{self.source!r}.{self.attribute_name}"


@dataclass(frozen=True)
class IndexArgumentRef(DelegatingArgumentRef):
    """ A reference to an element in a list. """

    index: int

    def map_value(self, value: Any, context: ErrorContext) -> Iterable[Tuple[Any, ErrorContext]]:
        return [(value[self.index], context)]

    def map_context(self, context: ErrorContext) -> ErrorContext:
        return StackContext(context, IndexContext(self.index))

    def __repr__(self) -> str:
        return f"{self.source!r}[{self.index}]"


@dataclass(frozen=True)
class AllElementsRef(DelegatingArgumentRef):
    """ A reference to all elements in a collection. """

    def map_value(self, value: Any, context: ErrorContext) -> Iterable[Tuple[Any, ErrorContext]]:
        return [(v, StackContext(context, IndexContext(i))) for i, v in enumerate(value)]

    def __repr__(self) -> str:
        return f"{self.source!r}[all]"


@dataclass(frozen=True)
class KeysRef(DelegatingArgumentRef):
    """ A reference to all keys of a mapping. """

    def map_value(self, value: Any, context: ErrorContext) -> Iterable[Tuple[Any, ErrorContext]]:
        return [(k, StackContext(context, MappingKeyContext(k))) for k in value]

    def __repr__(self) -> str:
        return f"{self.source!r}.keys()"


@dataclass(frozen=True)
class ValuesRef(DelegatingArgumentRef):
    """ A reference to all values of a mapping. """

    def map_value(self, value: Any, context: ErrorContext) -> Iterable[Tuple[Any, ErrorContext]]:
        return [(v, StackContext(context, MappingValueContext(k))) for k, v in value.items()]

    def __repr__(self) -> str:
        return f"{self.source!r}.values()"



# Copyright 2022 The GPflow Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=import-outside-toplevel

from typing import Any

from ..base_types import Shape
from ..error_contexts import ErrorContext
from ..shapes import register_get_shape


def install_tfp_integration() -> bool:
    """
    Install various hooks to support TensorFlow-Probability.

    If TensorFlow-Probability is not installed in this environment this function does nothing.

    :return: Whether TensorFlow-Probability support hooks actually were installed.
    """
    try:
        import tensorflow as tf
        import tensorflow_probability as tfp
    except ImportError:
        return False  # TensorFlow-Probability not installed - don't install integrations.

    @register_get_shape(tfp.util.DeferredTensor)
    def get_tensorflow_shape(shaped: Any, context: ErrorContext) -> Shape:
        shape = shaped.shape
        if not shape:
            return None
        return tuple(shape)

    # pylint: disable=protected-access  # To access the _TensorCoercible

    @register_get_shape(tfp.python.layers.internal.distribution_tensor_coercible._TensorCoercible)
    def get_tensor_coercible_shape(shaped: Any, context: ErrorContext) -> Shape:
        # This one is unpleasant. Sometimes the `shape` is a `TensorShape`, but sometimes it's a
        # function that returns a `TensorShape`. The version of TensorFlow probability seems to have
        # something to do with it, but it also seems to be more complicated...
        shape = shaped.shape
        if not shape:
            return None
        if not isinstance(shape, tf.TensorShape):
            shape = shape()
            if not shape:
                return None
        return tuple(shape)

    return True
